/*
 *
 * Copyright 2009-2011 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Suite 500, Boston, MA 02110-1335 USA.
 *
 */


#include <linux/kernel.h>
#include <linux/blkdev.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/version.h>
#include <linux/hdreg.h>
#include <linux/string.h>

/*
 * macro
 */
//#define NE3_RD_DEBUG

#ifdef NE3_RD_DEBUG
#define DBGPRINT    printk
#else
#define DBGPRINT(s,...)
#endif /* NE3_RD_DEBUG */

#define KERNEL_SECTOR_SIZE 512
#define SECTOR_SIZE_512    KERNEL_SECTOR_SIZE
#define SECTOR_SIZE_1024   (KERNEL_SECTOR_SIZE << 1)
#define SECTOR_SIZE_2048   (KERNEL_SECTOR_SIZE << 2)
#define SECTOR_SIZE_4096   (KERNEL_SECTOR_SIZE << 3)

/*
 * prototype
 */
static int __init ne3_rd_init(void);
static void __exit ne3_rd_exit(void);
static void ne3_rd_memfree(void);
static void ne3_rd_request(struct request_queue *queue);
int ne3_getgeo(struct block_device *block_device, struct hd_geometry *geo);

/*
 * global
 */
struct gendisk *ne3_rd_gd;
static struct request_queue *ne3_rd_queue;
static spinlock_t ne3_rd_lock;
static unsigned char* buffer;

/*
 * parameter
 */
static unsigned long debug = 0;  /* != 0 is debug mode. */
static unsigned short sector_size = 512;
static unsigned long sector_num = 2048;
static int major = 0;
static unsigned long minor_num_max = 16;
static char *name = "ne3_rd";
static unsigned long mem_start_addr = 0x60000000;
static unsigned long buf_mode = 1;  /* 0=PHY ADDR Direct 0!=vmalloc */

module_param(debug, ulong, S_IRUGO);
MODULE_PARM_DESC(debug, "Debug Print On/Off. 0 = off.");
module_param(sector_size, ushort, S_IRUGO);
MODULE_PARM_DESC(sector_size, "Sector Size");
module_param(sector_num, ulong, S_IRUGO);
MODULE_PARM_DESC(sector_num, "Sector Num");
module_param(major, int, S_IRUGO);
MODULE_PARM_DESC(major, "Major ID");
module_param(minor_num_max, ulong, S_IRUGO);
MODULE_PARM_DESC(minor_num_max, "MinorID Max Num");
module_param(name, charp, S_IRUGO);
MODULE_PARM_DESC(name, "Driver & Disk Name");
module_param(mem_start_addr, ulong, S_IRUGO);
MODULE_PARM_DESC(mem_start_addr, "RAM Start Addr(PHY Addr)");
module_param(buf_mode, ulong, S_IRUGO);
MODULE_PARM_DESC(buf_mode, "Buffer Mode 0=PHY ADDR Direct 0!=vmalloc");

/*
 * ops
 */
static struct block_device_operations ne3_rd_fops = {
    .owner   = THIS_MODULE,
    .getgeo = ne3_getgeo,
};


/*---------------------------------------------------------
    Init
----------------------------------------------------------*/
static int __init ne3_rd_init(void)
{
    int ret;
    int len;

    DBGPRINT(KERN_INFO "[ne3rd]============ne3_rd start==================\n");

    if ((sector_size != SECTOR_SIZE_512) &&
        (sector_size != SECTOR_SIZE_1024) &&
        (sector_size != SECTOR_SIZE_2048) &&
        (sector_size != SECTOR_SIZE_4096)) {
        DBGPRINT(KERN_WARNING "[ne3rd]sector_size must be 512, 1024, 2048 or 4096 byte\n");
        return -EINVAL;
    }

    if (((unsigned long long)sector_num * sector_size) > ULONG_MAX) {
        DBGPRINT(KERN_WARNING "[ne3rd]allocate size is over ULONG_MAX (sector_num * sector_size = 0x%llX)\n",
                              (unsigned long long)sector_num * sector_size);
        return -EINVAL;
    }

    if (minor_num_max < 1) {
        DBGPRINT(KERN_WARNING "[ne3rd]minor_num_max is smaller than 1\n");
        return -EINVAL;
    }

    if (name == NULL) {
        DBGPRINT(KERN_WARNING "[ne3rd]name is null\n");
        return -EINVAL;
    }

    len = strnlen(name, 16);
    if ((len == 0) || (len >= 16)) {
        DBGPRINT(KERN_WARNING "[ne3rd]length of name must be more than 1 character and the 15 characters following\n");
        return -EINVAL;
    }
    
    if (major < 0) {
        major = 0;
    }
    DBGPRINT(KERN_INFO "[ne3rd][debug] major=%d\n", major);

    /* Get Buffer */
    if(!buf_mode) {
        if ((mem_start_addr & (PAGE_SIZE-1)) != 0) {
            DBGPRINT(KERN_WARNING "[ne3rd]mem_start_addr is not PAGE_SIZE align\n");
            return -EINVAL;
        }
        if((buffer = (unsigned char *)ioremap(mem_start_addr, sector_num * sector_size)) == NULL) {
            return -ENOMEM;
        }
        DBGPRINT(KERN_INFO "[ne3rd]ioremap  buf_mode=%ld mem_start_addr=%x\n",
                                      buf_mode, (unsigned int)mem_start_addr);
    }
    else {
        if((buffer = (unsigned char *)vmalloc(sector_num * sector_size)) == NULL) {
            return -ENOMEM;
        }
        DBGPRINT(KERN_INFO "[ne3rd]vmalloc  buf_mode=%ld\n", buf_mode);
    }


    /* block dev queue init*/
    spin_lock_init(&ne3_rd_lock);

    if(!(ne3_rd_queue = blk_init_queue(ne3_rd_request, &ne3_rd_lock))) {
        DBGPRINT(KERN_WARNING "[ne3rd]blk_init_queue failed\n");                
        ne3_rd_memfree();
        return -ENOMEM;
    }

    /* set hardware sector size */
    blk_queue_logical_block_size(ne3_rd_queue, sector_size);

    /* Block Device Regist */
    ret = register_blkdev(major, name);
    if(ret < 0) {
        DBGPRINT(KERN_WARNING "[ne3rd]register_blkdev failed %x (%d)\n", major, ret);
        blk_cleanup_queue(ne3_rd_queue);
        ne3_rd_memfree();
        return -EBUSY;
    }
    if (major == 0) {
        major = ret;
    }

    /* alloc disk */
    if(!(ne3_rd_gd = alloc_disk(minor_num_max))) {
        DBGPRINT(KERN_WARNING "[ne3rd]alloc_disk failed\n");        
        blk_cleanup_queue(ne3_rd_queue);
        unregister_blkdev(major, name);
        ne3_rd_memfree();
        return -ENOMEM;
    }
    
    ne3_rd_gd->major = major;
    ne3_rd_gd->first_minor = 0;
    ne3_rd_gd->fops = &ne3_rd_fops;
    ne3_rd_gd->queue = ne3_rd_queue;
    strlcpy(ne3_rd_gd->disk_name, name, sizeof(ne3_rd_gd->disk_name));

    set_capacity(ne3_rd_gd, sector_num * (sector_size / KERNEL_SECTOR_SIZE));
    add_disk(ne3_rd_gd);

    DBGPRINT(KERN_INFO "[ne3rd][debug]major=%d\n", major);
    DBGPRINT(KERN_INFO "[ne3rd][debug]name=%s\n", name);
    DBGPRINT(KERN_INFO "[ne3rd][debug]sector_num=%ld\n", sector_num);
    DBGPRINT(KERN_INFO "[ne3rd][debug]sector_size=%d\n", sector_size);
    DBGPRINT(KERN_INFO "[ne3rd][debug]minor_num_max=%ld\n", minor_num_max);
    DBGPRINT(KERN_INFO "[ne3rd][debug]buf_mode=%ld\n", buf_mode);
    DBGPRINT(KERN_INFO "[ne3rd][debug]buffer=%x\n", (unsigned int)buffer);
    DBGPRINT(KERN_INFO "[ne3rd]Load Completed\n");
    
    return 0;
}


/*---------------------------------------------------------
    Exit
----------------------------------------------------------*/
static void __exit ne3_rd_exit(void)
{
    del_gendisk(ne3_rd_gd);
    put_disk(ne3_rd_gd);
    unregister_blkdev(major, name);
    blk_cleanup_queue(ne3_rd_queue);

    ne3_rd_memfree();

    DBGPRINT(KERN_INFO "[ne3rd]Unload Completed %d\n", major);
    DBGPRINT(KERN_INFO "[ne3rd]============ne3_rd end==================\n");

    return;
}

/*---------------------------------------------------------
    memory buffer free
----------------------------------------------------------*/
static void ne3_rd_memfree(void)
{
    if(!buf_mode) {
        iounmap(buffer);
    }
    else {
        vfree(buffer);
    }

    return;
}

/*---------------------------------------------------------
    Request
----------------------------------------------------------*/
static void ne3_rd_request(struct request_queue *queue)
{
    struct request *req;
    unsigned int offset, len;

    DBGPRINT(KERN_INFO "ne3_rd: %s\n", __FUNCTION__);

    req = blk_fetch_request(queue);
    while(req != NULL) {

        if(!blk_fs_request(req)) {
            printk(KERN_NOTICE "[ne3rd]blk_fs_request\n");
            __blk_end_request_all(req, -EIO);
            continue;
        }

        offset = blk_rq_pos(req) * KERNEL_SECTOR_SIZE;
        len    = blk_rq_cur_sectors(req) * KERNEL_SECTOR_SIZE;

        if((offset + len) > (sector_num * sector_size)) {
            printk( KERN_WARNING "[ne3rd]no buffer\n");
            return;
        }
        
        if(rq_data_dir(req)) {
            // write
            memcpy(buffer + offset, req->buffer, len);
        }
        else {
            // read
            memcpy(req->buffer, buffer + offset, len);
        }
        
        if(debug) {
            if(rq_data_dir(req)) {
                printk(KERN_INFO "[ne3rd]ne3_wr_request ");
            }
            else {
                printk(KERN_INFO "[ne3rd]ne3_rd_request ");
            }
            printk("blk_rq_pos(req)=%d ", (unsigned int)blk_rq_pos(req));
            printk("blk_rq_cur_sectors(req)=%d ", (unsigned int)blk_rq_cur_sectors(req));
            printk("buffer=%x ", (unsigned int)buffer);
            printk("offset=%u ", offset);
            printk("len=%u\n", len);
        }

        if ( ! __blk_end_request_cur(req, 0) ) {
            req = blk_fetch_request(queue);
        }
    }

    return;
}

/* heads       number of heads                    */
/* sectors     number of sectors/track            */
/* cylinders   number of cylinders, mod 65536     */
/* start       starting sector of this partition. */

int ne3_getgeo(struct block_device *block_device, struct hd_geometry *geo)
{
    unsigned long sectors_per_cylinder;

    /* Unused Parameter */
    (void)(block_device);

    DBGPRINT(KERN_INFO "ne3_rd: %s\n", __FUNCTION__);

    if (geo == NULL) {
        return -EINVAL;
    }

    /* We have no real geometry, of course, so make something up. */
    geo->heads = 4;
    geo->sectors = 16;
    geo->start = 0;
    sectors_per_cylinder = geo->heads * geo->sectors;
    geo->cylinders = (sector_num / sectors_per_cylinder) +
                                ((sector_num % sectors_per_cylinder) ? 1 : 0);

    if(debug) {
        printk("geo->heads    =%u\n", geo->heads);
        printk("geo->sectors  =%u\n", geo->sectors);
        printk("geo->start    =%lu\n", geo->start);
        printk("geo->cylinders=%u\n", geo->cylinders);
        printk("sectors_per_cylinder=%lu\n", sectors_per_cylinder);
    }

    return 0;
}

module_init(ne3_rd_init);
module_exit(ne3_rd_exit);

MODULE_DESCRIPTION("NE3 RAM Disk Driver");
MODULE_LICENSE("GPL v2");

